<?php
/**
 * Reads xdebug trace files and makes them simple to browse.
 * 
 * Usage:
 *  
 * 1. checkout trace reader into a directory on your webserver somewhere
 * 
 * 2. in config.php set TRACE_PATH to the absolute path where your xdebug 
 *    trace files live (this is the path used for xdebug.trace_output_dir)
 *    You may also want to change TRACE_FORMAT which is a regex to match
 * 
 * 3. create some xdebug traces by adding this to your server config:
 * 
 *    php_value xdebug.trace_output_dir /home/tomw/profiler
 *    php_value xdebug.trace_output_name trace.out.%t
 *    php_value xdebug.show_mem_delta 1
 *    php_value xdebug.trace_format 0
 *    php_value xdebug.auto_trace 1
 *    
 *    (see xdebug site for more details)
 *    
 * 4. open this tool in a browser to see the files and start working with them. 
 * 
 * Trace file format example:
 *   0.0002      71872   +71872   -> {main}() /usr/local/share/php/bootstrap.php:0
 * 
 * @author Tom Walter <walter.tom@gmail.com>
 */


require_once('config.php');

?>
<html>
<head>
<link rel="stylesheet" type="text/css" href="dataTables-1.4/media/css/demos.css"/>
<script language="javascript" src="jquery-1.3.2.min.js"></script>
<script language="javascript" src="dataTables-1.4/media/js/jquery.dataTables.min.js"></script>
<?php 

$fileName = null;
if (isset($_REQUEST['f'])) $fileName = $_REQUEST['f'];

if ($fileName === null)
{
	// list the available files
	?>
	<script language="javascript">
	$(document).ready(function() {
		$('#example').dataTable({
			"aaSorting": [[1,'desc']],
			"bProcessing": true, 
			"aoColumns": [
				null,
				{ "sType": "date" },
				null,
				null
			]
		});
	});
	</script>
	</head>
	<body>
	<h3>Trace files - (most recent first)</h3>
	<table id="example" class="display">
		<thead>
		<tr align="left">
			<th>Name</th>
			<th>Date</th>
			<th>Filesize</th>
			<th>Filesize (B)</th>
		</tr>
		</thead>
		<tbody>
		<?php 
			foreach (scandir(TRACE_PATH) as $traceFile)
			{
				if (preg_match(TRACE_FORMAT, $traceFile))
				{
					$traceFilePath = TRACE_PATH.$traceFile;
					?>
					<tr>
						<td><a href="?f=<?= $traceFile ?>"><?= $traceFile ?></a></td>
						<td><?= date(DATE_RFC1123, filemtime($traceFilePath)) ?></td>
						<td><?= memoryFormat(filesize($traceFilePath)) ?></td>
						<td><?= filesize($traceFilePath) ?></td>
					</tr>
					<?php 
				} 
			}
		?>
		</tbody>
	</table>
	</body>
	</html>
	<?php 
	exit;
}

$fileName = TRACE_PATH.$fileName;
if (!file_exists($fileName)) die("$fileName does not exist");

?>
<script language="javascript">
$(document).ready(function() {
	$('#example').dataTable({
		"aaSorting": [[6,'desc'], [4,'desc']],
		"bProcessing": true, 
		"aoColumns": [
			null,
			{ "bSortable": false },
			null,
			null,
			null,
			null,
			null
		]
	});
});
</script>
</head>
<body>
<?

error_reporting(E_ALL) ;
ini_set('display_errors', true);
putenv('LC_ALL=C'); // speeds up grep heaps

$startOffset = 0;
if (isset($_REQUEST['o'])) $startOffset = $_REQUEST['o'];

$records = array();

$offsetIdx = 0;
$timeIdx = 1;
$memIdx = 2;
$deltaIdx = 3;
$funcIdx = 4;
$pathIdx = 5;
$lineIdx = 6;


function processLine($line, $offset)
{
	// for ungrepped lines
	global $records;
	$fields = sscanf($line, '%f%u%d -> %s %[/A-Za-z0-9._-]:%u');
	if (count($fields))
	{
		array_unshift($fields, $offset);
		$records[] = $fields;
	}
}

function processGrepLine($line, $grepOffset)
{
	// for grepped lines which contain byte offset at start
	global $records, $startOffset;
	$fields = sscanf($line, '%u: %f%u%d -> %s %[/A-Za-z0-9._-]:%u');
	if (count($fields))
	{
		$fields[0] += $grepOffset-1;
		$records[] = $fields;
	}
}

function memoryFormat($bytes)
{
	$kb = $bytes/1024;
	if ($kb<1 && $kb>-1) return $bytes .' B';
	$mb = $kb/1024;
	if ($mb<1 && $mb>-1) return sprintf('%.2f KB', $kb);
	return sprintf('%.2f MB', $mb);
}

$file = fopen($fileName, 'r');
echo('<h3><a href="?f='.$_REQUEST['f'].'">'.$_REQUEST['f'].'</a> - "'.fgets($file).'"</h3>'); // header line
$getFinalLine = true;
if ($startOffset === 0)
{
	// initial load
	$offset = ftell($file);
	$firstLine = fgets($file);
	$levelIndent = strpos($firstLine, '->');
	processLine($firstLine, $offset);
	$offset = ftell($file);
	fclose($file);
	$stream = popen('tail -c +'.$offset.' '.$fileName.' | grep -b "^.\{'.$levelIndent.'\}->"', 'r');
	while (!feof($stream)) 
	{
		$line = fgets($stream);
   		processGrepLine($line, $offset);
	}
	pclose($stream); 
}
else
{
	fseek($file, $startOffset);
	$parentLine = fgets($file);
	echo('<p><strong>'.$parentLine.'</strong></p>');
	$childIndent = strpos($parentLine, '->') + 3;
	$grepOffset = ftell($file);
	$stream = popen('tail -c +'.$grepOffset.' '.$fileName.' | grep -b "^.\{1,'.$childIndent.'\}->"', 'r');
	fclose($file);
    while (!feof($stream)) 
    {
    	$line = fgets($stream);
    	$indent = strpos($line, '->') - strpos($line, ':');
    	if ($indent == $childIndent) 
    	{
    		processGrepLine($line, $grepOffset);
    	} 
    	else if ($indent !== 0 && $indent < $childIndent) 
    	{
    		$getFinalLine = false; // don't need to because we found a parent line
    		processGrepLine($line, $grepOffset);
    		break;
    	}
    }
    pclose($stream);
}
if ($getFinalLine)
{
	// go back and get the final trace line so we can calculate deltas
    $stream = popen('tail -n 4 '.$fileName, 'r');
    $line = fgets($stream);
    if ($line)
    	processLine($line, null);
    pclose($stream);
}

?>
<table id="example" class="display">
	<thead>
	<tr align="left">
		<th>Order</th>
		<th align="right">&raquo;</th>
		<th>Action</th>
		<th>File</th>
		<th>Time (s)</th>
		<th>Memory</th>
		<th>Mem (B)</th>
	</tr>
	</thead>
	<tbody>
<?php
$prevRecord = null;
foreach ($records as $idx=>$record)
{
	if ($prevRecord === null) 
	{
		$prevRecord = $record;
		continue; 
	}
	$mem = $record[$memIdx] - $prevRecord[$memIdx];
	$fullPath = $prevRecord[$pathIdx];
	$path = substr($fullPath, strrpos($fullPath, '/')+1);
	?>
	<tr>
		<td><?= $idx ?></td>
		<td align="right"><a href="?f=<?= $_REQUEST['f'] ?>&o=<?= $prevRecord[$offsetIdx] ?>">&raquo;</a></td>
		<td><?= $prevRecord[$funcIdx] ?></td>
		<td><a title="<?= $fullPath ?>" href="fileviewer.phtml?f=<?= $fullPath ?>&l=<?= $prevRecord[$lineIdx] ?>"><?= $path ?>:<?= $prevRecord[$lineIdx] ?></a></td>
		<td><?= $record[$timeIdx] - $prevRecord[$timeIdx] ?></td>
		<td><?= memoryFormat($mem) ?></td>
		<td><?= $mem ?></td>
	</tr>
	<?php
	$prevRecord = $record; 
} 
?>
	</tbody>
</table>
<p style="clear:both"><a href='?'>&laquo; Back to trace file list</a></p>
</body>
</html>